use serde_json;
use serde_json::json;
use std::error::Error;

extern crate custom_error;
use custom_error::custom_error;

const JSON_POINTER_SEPARATOR: &str = "/";

custom_error! { pub IncorrectPointer{pointer: String} = "Incorrect pointer is specified: {pointer}" }

enum ValueReference<'a> {
    Object(&'a mut serde_json::Map<String, serde_json::Value>, String),
    Array(&'a mut Vec<serde_json::Value>, usize),
    Invalid,
}

#[derive(Debug)]
pub struct File {
    contents: serde_json::Value,
}

impl ToString for File {
    fn to_string(&self) -> String {
        self.contents.to_string()
    }
}

impl File {
    pub fn empty() -> File {
        File {
            contents: json!({}),
        }
    }

    pub fn new(file_contents: String) -> Result<File, Box<dyn Error>> {
        Ok(File {
            contents: serde_json::from_str(&file_contents)?,
        })
    }

    pub fn root(&self) -> &serde_json::Value {
        &self.contents
    }

    pub fn get(&self, pointer: &str) -> Option<&serde_json::Value> {
        self.contents.pointer(pointer)
    }

    pub fn contains(&self, pointer: &str) -> bool {
        self.get(pointer) != None
    }

    pub fn insert(
        &mut self,
        pointer: &str,
        new_value: serde_json::Value,
    ) -> Result<(), IncorrectPointer> {
        self.touch(pointer)?;
        match self.contents.pointer_mut(pointer) {
            Some(v) => *v = new_value,
            _ => {
                //  If after `touch()` call we still don't have an existing value -
                //  something is wrong with the `pointer`
                return Err(IncorrectPointer {
                    pointer: pointer.to_owned(),
                });
            }
        };
        Ok(())
    }

    pub fn remove(&mut self, pointer: &str) -> Option<serde_json::Value> {
        match self.pointer_to_reference(pointer) {
            ValueReference::Object(map, variable_name) => map.remove(&variable_name),
            ValueReference::Array(vec, variable_index) => {
                if variable_index < vec.len() {
                    return Some(vec.remove(variable_index));
                }
                None
            }
            _ => None,
        }
    }

    fn touch(&mut self, pointer: &str) -> (Result<(), IncorrectPointer>) {
        //  If value is present - we're done
        if pointer.is_empty() || self.contents.pointer_mut(pointer).is_some() {
            return Ok(());
        }
        //  Otherwise - try to create it
        match self.pointer_to_reference(pointer) {
            ValueReference::Object(map, variable_name) => {
                map.insert(variable_name, json!(null));
            }
            ValueReference::Array(vec, variable_index) => {
                //  We've checked at the beginning of this method that value
                //  at `variable_index` does not exist, which guarantees
                //  that array is to short and we won't shrink it
                vec.resize(variable_index + 1, json!(null));
            }
            _ => {
                return Err(IncorrectPointer {
                    pointer: pointer.to_owned(),
                })
            }
        };
        Ok(())
    }

    fn pointer_to_reference<'a>(&'a mut self, pointer: &str) -> ValueReference<'a> {
        if pointer.is_empty() {
            return ValueReference::Invalid;
        }
        //  Extract variable name (that `pointer` points to)
        //  and reference to it's container
        //
        //  i.e.    given file with '{"obj":{"arr":[1,3,5,2,4]}}',
        //          for pointer `/obj/arr/5`,
        //  it will return, basically, `(&[1,3,5,2,4], "5")`
        let container_variable_pair =
            pop_json_pointer(pointer).and_then(move |(path, variable_name)| {
                match self.contents.pointer_mut(&path) {
                    Some(v) => Some((v, variable_name)),
                    _ => None,
                }
            });
        let (json_container, variable_name) = match container_variable_pair {
            Some(v) => v,
            _ => return ValueReference::Invalid,
        };
        //  For arrays we also need to confirm validity of the variable name
        //  and convert it into `usize`
        match json_container {
            serde_json::Value::Object(map) => ValueReference::Object(map, variable_name),
            serde_json::Value::Array(vec) => {
                let index: usize = match variable_name.parse() {
                    Ok(v) => v,
                    _ => return ValueReference::Invalid,
                };
                ValueReference::Array(vec, index)
            }
            _ => ValueReference::Invalid,
        }
    }
}

fn pop_json_pointer(pointer: &str) -> Option<(String, String)> {
    let mut pointer = pointer.to_string();
    let last_separator_index = match pointer.rfind(JSON_POINTER_SEPARATOR) {
        Some(v) => v,
        _ => {
            return None;
        }
    };
    if last_separator_index >= pointer.len() {
        pointer.pop();
        return Some((pointer, String::new()));
    }
    let var_name = pointer.split_off(last_separator_index + 1);
    pointer.pop();
    Some((pointer, var_name))
}

#[test]
fn test_pop_json_pointer() {
    assert_eq!(
        pop_json_pointer("/a/b/c/d"),
        Some(("/a/b/c".to_owned(), "d".to_owned()))
    );
    assert_eq!(pop_json_pointer("/"), Some(("".to_owned(), "".to_owned())));
    assert_eq!(
        pop_json_pointer("/a/b/"),
        Some(("/a/b".to_owned(), "".to_owned()))
    );
    assert_eq!(pop_json_pointer(""), None);
    //  This pointer is incorrect
    assert_eq!(pop_json_pointer("var"), None);
}
