#[cfg(test)]
mod tests;

use serde_json;
use std::collections::HashMap;
use std::error::Error;
use std::path;

extern crate custom_error;
use custom_error::custom_error;

pub mod file;
pub use file::File;

pub mod io;

custom_error! { pub DBError
    InvalidEntityName{entity_name: String} = r#"Cannot use {entity_name} for file or group"#,
    NoGroup{group_name: String} = r#"Group "{group_name}" does not exist"#,
    NoFile{group_name: String, file_name: String} = r#"There is no "{file_name}" file in group "{group_name}""#,
    GroupAlreadyExists{group_name: String} = r#"Group "{group_name}" already exists"#,
    FileAlreadyExists{group_name: String, file_name: String} = r#"File "{file_name}" already exists in the group "{group_name}""#,
}

pub struct Group {
    name: String,
    files: HashMap<String, File>,
}

pub struct Database {
    storage_path: path::PathBuf,
    groups: Vec<Group>,
}

impl Database {
    pub fn new(storage_path: &path::Path) -> Result<Database, Box<dyn Error>> {
        Ok(Database {
            storage_path: storage_path.to_path_buf(),
            groups: io::read(storage_path)?,
        })
    }

    pub fn clear(&mut self) {
        self.groups = Vec::new();
    }

    pub fn path(&mut self) -> path::PathBuf {
        self.storage_path.clone()
    }

    pub fn change_path(&mut self, new_path: &path::Path) -> Result<(), Box<dyn Error>> {
        self.write_copy(new_path)?;
        io::clear_dir(&self.storage_path)?;
        self.storage_path = new_path.to_path_buf();
        Ok(())
    }

    pub fn write_copy(&self, new_path: &path::Path) -> Result<(), Box<dyn Error>> {
        io::clear_dir(new_path)?;
        io::write(new_path, &self)?;
        Ok(())
    }

    pub fn erase(self) -> Result<(), Box<dyn Error>> {
        io::clear_dir(&self.storage_path)?;
        Ok(())
    }

    pub fn group_names(&self) -> Vec<String> {
        self.groups.iter().map(|x| x.name.clone()).collect()
    }

    pub fn file_names_in(&self, group_name: &str) -> Vec<String> {
        match self.groups.iter().find(|x| x.name.eq(group_name)) {
            Some(group) => group.files.keys().map(|x| x.clone()).collect(),
            None => Vec::new(),
        }
    }

    pub fn contains_group(&self, group_name: &str) -> bool {
        self.group_index(group_name).is_ok()
    }

    pub fn create_group(&mut self, group_name: &str) -> Result<(), DBError> {
        assert_name_is_valid(group_name)?;
        self.assert_no_group(group_name)?;
        self.groups.push(Group {
            name: group_name.to_owned(),
            files: HashMap::new(),
        });
        Ok(())
    }

    pub fn remove_group(&mut self, group_name: &str) -> Result<(), DBError> {
        let index = self.group_index(group_name)?;
        self.groups.remove(index);
        Ok(())
    }

    pub fn contains_file(&self, group_name: &str, file_name: &str) -> bool {
        self.group_files(group_name)
            .and_then(|x| Ok(x.contains_key(&file_name.to_owned())))
            .unwrap_or(false)
    }

    pub fn create_file(&mut self, group_name: &str, file_name: &str) -> Result<&mut File, DBError> {
        assert_name_is_valid(&file_name)?;
        let files = self.group_files_mut(group_name)?;
        if files.contains_key(&file_name.to_owned()) {
            return Err(DBError::FileAlreadyExists {
                group_name: group_name.to_owned(),
                file_name: file_name.to_owned(),
            });
        }
        let new_file = File::empty();
        files.insert(file_name.to_owned(), new_file);
        Ok(files
            .get_mut(file_name)
            .expect("Missing value that was just inserted."))
    }

    pub fn remove_file(&mut self, group_name: &str, file_name: &str) -> Result<(), DBError> {
        if self
            .group_files_mut(group_name)?
            .remove(file_name)
            .is_none()
        {
            return Err(DBError::NoFile {
                group_name: group_name.to_owned(),
                file_name: file_name.to_owned(),
            });
        }
        Ok(())
    }

    pub fn file_mut(&mut self, group_name: &str, file_name: &str) -> Option<&mut File> {
        match self.group_files_mut(group_name) {
            Ok(files) => files.get_mut(&file_name.to_owned()),
            _ => None,
        }
    }

    pub fn file(&self, group_name: &str, file_name: &str) -> Option<&File> {
        match self.group_files(group_name) {
            Ok(files) => files.get(&file_name.to_owned()),
            _ => None,
        }
    }

    fn assert_no_group(&self, group_name: &str) -> Result<(), DBError> {
        if self.group_index(group_name).is_ok() {
            return Err(DBError::GroupAlreadyExists {
                group_name: group_name.to_owned(),
            });
        }
        Ok(())
    }

    fn group_index(&self, group_name: &str) -> Result<usize, DBError> {
        match self.groups.iter().position(|x| x.name.eq(group_name)) {
            Some(index) => Ok(index),
            _ => {
                return Err(DBError::NoGroup {
                    group_name: group_name.to_owned(),
                })
            }
        }
    }

    fn group_files_mut(&mut self, group_name: &str) -> Result<&mut HashMap<String, File>, DBError> {
        let group_index = self.group_index(group_name)?;
        Ok(&mut (&mut self.groups[group_index]).files)
    }

    fn group_files(&self, group_name: &str) -> Result<&HashMap<String, File>, DBError> {
        let group_index = self.group_index(group_name)?;
        Ok(&self.groups[group_index].files)
    }
}

fn is_name_valid(entity_name: &str) -> bool {
    entity_name
        .chars()
        .all(|x| x.is_ascii_alphabetic() || x.is_ascii_digit())
}

fn assert_name_is_valid(entity_name: &str) -> Result<(), DBError> {
    if is_name_valid(entity_name) {
        return Ok(());
    }
    return Err(DBError::InvalidEntityName {
        entity_name: entity_name.to_owned(),
    });
}

// TODO make sure file's main value not being an object won't break anything
// TODO handle parsing errors differently
// TODO add logs
// TODO add docs
